#include <proto.h>
#include <sys/stat.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
// Stupid default totaly unportable probably.
#define O_RDONLY	0
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_FNMATCH
#include <fnmatch.h>
#endif

#include "gtfilereq.h"
#include "gtgroup.h"
#include "gtbutton.h"
#include "gtlabel.h"
#include "gtspace.h"
#include "gtseparator.h"
#include "gtapp.h"
#include "gterror.h"
#include "gtcommon.h"
#include "gtdlink.h"

/************************************************************************\
*
* GtFileReq class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

/* Defines to handle directory scanning. */
#if HAVE_DIRENT_H
#include <dirent.h>
#define NAMELEN(dirent) strlen((dirent)->d_name)
#else
#define dirent direct
#define NAMELEN(dirent) (dirent)->d_namlen
#if HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#if HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#if HAVE_NDIR_H
#include <ndir.h>
#endif
#endif

/* Type of directory entries */
typedef struct GtFREntry
{
	GtListItem li;
	sbool dir;
	char name[1];
} GtFREntry;

/************************************************************************\
*
* Create and init filereq.
*
\************************************************************************/
GtFileReq *GtMkFileReq(char *n,GtManager *p,GtTag *tgs)
{
	GtFileReq *obj;
	
	if(!(obj=new GtFileReq()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtFileReq object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init filereq.
*
\************************************************************************/
GtFileReq *GtMkFileReq(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtFileReq *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkFileReq(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}

/************************************************************************\
*
* Setup empty filereq.
*
\************************************************************************/
GtFileReq::GtFileReq(void)
{
	grp=NULL;
	load=NULL;
	save=NULL;
	sl=NULL;
	pat=NULL;
	file=NULL;
	home = -1;
	curr = -1;
}

/************************************************************************\
*
* Destroy filereq.
*
\************************************************************************/
void GtFileReq::Destroy(void)
{
	if(grp)
		GtDestroy(grp);
	if(curr >= 0)
		close(curr);
	if(home >= 0)
	{
		fchdir(home);
		close(home);
	}
	GtWidget::Destroy();
}

/************************************************************************\
*
* Setup filereq.
*
\************************************************************************/
sbool GtFileReq::Init(char *n,GtManager *pr,GtTag *tgs)
{
	GtTag *t;
	GtList *lst;
	GtButton *b;
	char *p="*",*d=".",*f=NULL;
	sbool ld=TRUE;

	if(!GtWidget::Init(n,pr,tgs))
		return(FALSE);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Pattern:
				p=(char *)t->d.ptr;
				break;
			case GT_Directory:
				d=(char *)t->d.ptr;
				break;
			case GT_File:
				f=(char *)t->d.ptr;
				break;
			case GT_LoadSave:
				if(t->d.val)
					ld=TRUE;
				else
					ld=FALSE;
				break;
		}
	}
	if(!(grp=GtMkGroup("frgroup",pr,GT_Orient,GTO_VERT,GT_DONE)))
		return(FALSE);
	if(!(sl=GtMkScrList("slist",grp,NULL)))
		return(FALSE);
	lst=sl->GetList();
	lst->SetV(GT_Mode,GTLM_ALWAYSONE,GT_DONE);
	lst->AddCB(GT_Select,ListSelect,this);
	if(!GtMkLabel("pattern",grp,
			GT_Weight,0,
			GT_Place,GTP_TOPLEFT,
			#ifndef HAVE_FNMATCH
			GT_Mapped,FALSE,
			#endif
			GT_DONE))
		return(FALSE);
	if(!(pat=GtMkStrInput("patstr",grp,
			GT_MaxLength,100,
			GT_Weight,0,
			#ifndef HAVE_FNMATCH
			GT_Mapped,FALSE,
			#endif
			GT_DONE)))
		return(FALSE);
	pat->AddCB(GT_Done,PatternDone,this);
	if(!GtMkLabel("file",grp,
			GT_Weight,0,
			GT_Place,GTP_TOPLEFT,
			GT_DONE))
		return(FALSE);
	if(!(file=GtMkStrInput("filestr",grp,
			GT_MaxLength,256,
			GT_Weight,0,
			GT_DONE)))
		return(FALSE);
	file->AddCB(GT_Done,FileDone,this);
	if(!GtMkSeparator("sep",grp,GT_Orient,GTO_HORIZ,GT_Weight,0,GT_DONE))
		return(FALSE);
	if(!(bgrp=GtMkGroup("bgroup",grp,GT_Weight,0,GT_DONE)))
		return(FALSE);
	if(!GtMkSpace("space",bgrp,NULL))
		return(FALSE);
	if(!(load=GtMkButton("loadbut",bgrp,GT_Mapped,(ld ? TRUE : FALSE),GT_DONE)))
		return(FALSE);
	if(!GtMkLabel("load",load,NULL))
		return(FALSE);
	load->AddCB(GT_Activate,LoadSave,this);
	if(!(save=GtMkButton("savebut",bgrp,GT_Mapped,(ld ? FALSE : TRUE),GT_DONE)))
		return(FALSE);
	if(!GtMkLabel("save",save,NULL))
		return(FALSE);
	save->AddCB(GT_Activate,LoadSave,this);
	if(!GtMkSpace("space",bgrp,NULL))
		return(FALSE);
	if(!(b=GtMkButton("cancelbut",bgrp,NULL)))
		return(FALSE);
	if(!GtMkLabel("cancel",b,NULL))
		return(FALSE);
	b->AddCB(GT_Activate,Cancel,this);
	if(!GtMkSpace("space",bgrp,NULL))
		return(FALSE);
	if(p)
		pat->ReplaceString(p);
	if(f)
		file->ReplaceString(f);
	home = open(".",O_RDONLY);
	if(home < 0)
	{
		GtError(this,GTEC_MISC,"Unable to open original directory!\n");
		return(FALSE);
	}
	curr = open(d,O_RDONLY);
	if(curr < 0)
		curr = home;
	return(TRUE);
}

/************************************************************************\
*
* Set filereq attributes.
*
\************************************************************************/
sbool GtFileReq::Set(GtTag *tgs)
{
	GtTag *t;
	char *p=NULL,*d=NULL,*f=NULL;

	if(!GtWidget::Set(tgs))
		return(FALSE);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Pattern:
				p=(char *)t->d.ptr;
				break;
			case GT_Directory:
				d=(char *)t->d.ptr;
				break;
			case GT_File:
				f=(char *)t->d.ptr;
				break;
			case GT_LoadSave:
				if(t->d.val)
				{
					save->Map(FALSE);
					load->Map(TRUE);
				}
				else
				{
					load->Map(FALSE);
					save->Map(TRUE);
				}
				break;
		}
	}
	this->SetPattern(p);
	this->SetDir(d);
	this->SetFile(f);
	return(TRUE);
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtFileReq::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtFileReq")==0)
		return(TRUE);
	else
		return(GtWidget::IsSubClassOf(cn));
}

/************************************************************************\
*
* Set pattern for directory scan.
*
\************************************************************************/
void GtFileReq::SetPattern(char *p)
{
	if(p)
	{
		pat->ReplaceString(p);
		this->ScanDir();
	}
}

/************************************************************************\
*
* Set directory for scan.
*
\************************************************************************/
void GtFileReq::SetDir(char *d)
{
	int nc;
	
	if(d)
	{
		nc = open(d,O_RDONLY);
		if(nc >= 0)
		{
			if(curr >= 0)
				close(curr);
			curr = nc;
		}
		this->ScanDir();
	}
}

/************************************************************************\
*
* Set file to load/save.
*
\************************************************************************/
void GtFileReq::SetFile(char *f)
{
	if(f)
		file->ReplaceString(f);
}

/************************************************************************\
*
* Add filereq callback.
*
\************************************************************************/
void GtFileReq::AddCB(ulong tag,GtCallB cb,GtUDPtr ud)
{
	if(tag==GT_Done)
		done.AddCB(cb,ud);
	else
		GtWidget::AddCB(tag,cb,ud);
}

/************************************************************************\
*
* Remove filereq callback.
*
\************************************************************************/
void GtFileReq::RemCB(ulong tag,GtCallB cb)
{
	if(tag==GT_Done)
		done.RemCB(cb);
	else
		GtWidget::RemCB(tag,cb);
}

/************************************************************************\
*
* Startup filereq.
*
\************************************************************************/
void GtFileReq::Activate(void)
{
	this->ScanDir();
	file->Activate();
}

/************************************************************************\
*
* Deactivate filereq and restore current directory.
*
\************************************************************************/
void GtFileReq::Deactivate(void)
{
	fchdir(home);
	if(curr >= 0)
		close(curr);
	curr = -1;
}

/************************************************************************\
*
* Scan dir and add files to list.
*
\************************************************************************/
void GtFileReq::ScanDir(void)
{
	GtDLink hd;
	GtFREntry *fe,*tmp;
	char *p;
	int elen;
	DIR *dh;
	struct dirent *de;
	struct stat st;
	ulong filecol,dircol;
	
	sl->GetList()->GetV(GT_Foreground,&filecol,GT_DONE);
	dircol = filecol;
	GtDInit(&hd);
	p=pat->GetString();
	fchdir(curr);
	if((dh=opendir(".")))
	{
		while((de=readdir(dh)))
		{
			elen=NAMELEN(de);
			stat(de->d_name,&st);
			#ifdef HAVE_FNMATCH
			if((strcmp(de->d_name,".")==0) || ((fnmatch(p,de->d_name,0)!=0) &&
				!S_ISDIR(st.st_mode)))
			#else
			if((strcmp(de->d_name,".")==0) && !S_ISDIR(st.st_mode))
			#endif
				continue;
			if(S_ISDIR(st.st_mode))
				elen+=6;
			if(!(fe=(GtFREntry *)malloc(sizeof(GtFREntry)+elen)))
			{
				GtError(this,GTEC_NOMEMORY,"No memory for file list!\n");
				return;
			}
			if(S_ISDIR(st.st_mode))
			{
				strcpy(fe->name,"(dir) ");
				strcat(fe->name,de->d_name);
				fe->dir=TRUE;
				fe->li.color=dircol;
			}
			else
			{
				strcpy(fe->name,de->d_name);
				fe->dir=FALSE;
				fe->li.color=filecol;
			}
			fe->li.str=fe->name;
			fe->li.len=strlen(fe->name);
			fe->li.sel=FALSE;
			for(tmp=(GtFREntry *)GtDFirst(&hd); tmp;
					tmp=(GtFREntry *)GtDNext(&hd,&tmp->li.lnk))
			{
				if(fe->dir)
				{
					if(!tmp->dir || (strcmp(fe->name,tmp->name)<0))
					{
						GtDAddBefore(&tmp->li.lnk,&fe->li.lnk);
						break;
					}
				}
				else
				{
					if(!tmp->dir && (strcmp(fe->name,tmp->name)<0))
					{
						GtDAddBefore(&tmp->li.lnk,&fe->li.lnk);
						break;
					}
				}
			}
			if(!tmp)
				GtDAddTail(&hd,&fe->li.lnk);
		}
		sl->GetList()->ClrListItems();
		sl->GetList()->AddItems(&hd,0);
	}
}

/************************************************************************\
*
* List selection of file or directory.
*
\************************************************************************/
int GtFileReq::ListSelect(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtFileReqCD fcd;
	GtFREntry *fe;
	GtListSelectCD *lcd;
	GtFileReq *fr;
	
	obj = obj;
	lcd = (GtListSelectCD *)cd;
	fr = (GtFileReq *)ud;
	fe = (GtFREntry *)lcd->item;
	if(lcd->multi >= 2)
	{
		fcd.file = fr->file->GetString();
		fcd.ok = TRUE;
		fr->done.CallCB(fr,&fcd);
	}
	else
	{
		if(fe->dir && (lcd->multi != 0))
			fr->SetDir(&fe->name[6]);
		else
			fr->file->ReplaceString(fe->name);
	}
	return(GTCB_CONT);
}

/************************************************************************\
*
* Done editing pattern string.
*
\************************************************************************/
int GtFileReq::PatternDone(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtFileReq *fr;
	GtSIDoneCD *scd;
	
	obj = obj;
	scd = (GtSIDoneCD *)cd;
	fr = (GtFileReq *)ud;
	if(scd->why != GTDW_MOUSE)
		fr->file->Activate();
	fr->ScanDir();
	return(GTCB_CONT);
}

/************************************************************************\
*
* Done editing pattern string.
*
\************************************************************************/
int GtFileReq::FileDone(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtFileReqCD fcd;
	GtFileReq *fr;
	GtSIDoneCD *scd;
	GtListItem *l;
	GtList *lst;
	GtFREntry *fe;
	struct stat st;
	
	obj = obj;
	scd = (GtSIDoneCD *)cd;
	fr = (GtFileReq *)ud;
	lst = fr->sl->GetList();
	switch(scd->why)
	{
		case GTDW_UPTAB:
		case GTDW_DOWNTAB:
			fr->pat->Activate();
			break;
		case GTDW_UP:
			l=lst->GetNextPrevItem(lst->GetSelected(NULL),FALSE);
			if(l)
			{
				lst->SelectItem(l);
				lst->ShowItem(l);
				fe=(GtFREntry *)l;
				if(fe->dir)
					fr->file->ReplaceString(&fe->name[6]);
				else
					fr->file->ReplaceString(fe->name);
			}
			fr->file->Activate();
			break;
		case GTDW_DOWN:
			l=lst->GetNextPrevItem(lst->GetSelected(NULL),TRUE);
			if(l)
			{
				lst->SelectItem(l);
				lst->ShowItem(l);
				fe=(GtFREntry *)l;
				if(fe->dir)
					fr->file->ReplaceString(&fe->name[6]);
				else
					fr->file->ReplaceString(fe->name);
			}
			fr->file->Activate();
			break;
		case GTDW_ENTER:
			if(scd->len)
			{
				if((stat(scd->str,&st) >= 0) && S_ISDIR(st.st_mode))
				{
					fr->SetDir(scd->str);
					fr->file->Activate();
				}
				else
				{
					fcd.file = fr->file->GetString();
					fcd.ok = TRUE;
					fr->done.CallCB(fr,&fcd);
				}
			}
			else
			{
				fcd.file = NULL;
				fcd.ok = FALSE;
				fr->done.CallCB(fr,&fcd);
			}
			break;
	}
	return(GTCB_CONT);
}

/************************************************************************\
*
* Load / Save button pressed.
*
\************************************************************************/
int GtFileReq::LoadSave(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtFileReqCD fcd;
	GtFileReq *fr;
	
	obj = obj;
	cd = cd;
	fr = (GtFileReq *)ud;
	fcd.file = fr->file->GetString();
	fcd.ok = TRUE;
	fr->done.CallCB(fr,&fcd);
	return(GTCB_CONT);
}

/************************************************************************\
*
* Cancel button pressed.
*
\************************************************************************/
int GtFileReq::Cancel(GtBase *obj,GtCDPtr cd,GtUDPtr ud)
{
	GtFileReqCD fcd;
	GtFileReq *fr;
	
	obj = obj;
	cd = cd;
	fr = (GtFileReq *)ud;
	fcd.file = NULL;
	fcd.ok = FALSE;
	fr->done.CallCB(fr,&fcd);
	return(GTCB_CONT);
}
