#include <proto.h>

#include "gtmenu.h"
#include "gterror.h"
#include "gtcommon.h"

/************************************************************************\
*
* GtMenu class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

/************************************************************************\
*
* Create and init menu.
*
\************************************************************************/
GtMenu *GtMkMenu(char *n,GtManager *p,GtTag *tgs)
{
	GtMenu *obj;
	
	if(!(obj=new GtMenu()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtMenu object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init menu.
*
\************************************************************************/
GtMenu *GtMkMenu(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtMenu *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkMenu(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}

/************************************************************************\
*
* Setup empty menu.
*
\************************************************************************/
GtMenu::GtMenu(void)
{
	frm=NULL;
	sub=NULL;
	but=0;
}

/************************************************************************\
*
* Setup menu.
*
\************************************************************************/
sbool GtMenu::Init(char *n,GtManager *p,GtTag *tgs)
{
	if(!GtPopup::Init(n,p,tgs))
		return(FALSE);
	this->Map(FALSE);
	return(TRUE);
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtMenu::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtMenu")==0)
		return(TRUE);
	else
		return(GtPopup::IsSubClassOf(cn));
}

/************************************************************************\
*
* Build menu window and set saveunder attribute.
*
\************************************************************************/
sbool GtMenu::Build(void)
{
	if(flags & GTWF_UNBUILT)
		gtwin->SetV(GT_SaveUnder,TRUE,GT_DONE);
	return(GtPopup::Build());
}

/************************************************************************\
*
* Handle event and wait for button up.
*
\************************************************************************/
int GtMenu::Handle(XEvent *xe)
{
	int ret;

	if(!(flags & GTWF_MAPPED))
		return(GTHR_SEARCH);
	ret=GtPopup::Handle(xe);
	if(xe && (xe->type==ButtonRelease) && (xe->xbutton.button==but))
		this->Popdown();
	if((xe->type!=Expose) && dsp->HasGrab(this))
		return(GTHR_USED);
	return(ret);
}

/************************************************************************\
*
* Set menu event mask.
*
\************************************************************************/
void GtMenu::SetEMask(ulong em)
{
	GtPopup::SetEMask(em|ButtonReleaseMask|EnterWindowMask);
}

/************************************************************************\
*
* Notify popdown of sub menu.
*
\************************************************************************/
void GtMenu::NotifyA(ulong tag,void *arg)
{
	if(tag==GT_Popdown)
		this->PopdownSub((Time)arg);
	else
		GtPopup::NotifyA(tag,arg);
}
	
/************************************************************************\
*
* Popup menu from other object.
*
\************************************************************************/
void GtMenu::Popup(GtWidget *from,ulong button,Time tm)
{
	long dir=GTD_DOWN;
	int xp,yp,xs,ys;
	Window dmy,dmy2;
	unsigned int dmsk;
	
	if(flags & GTWF_UNBUILT)
		return;
	but=button;
	frm=from;
	if(frm)
	{
		XTranslateCoordinates(dsp->XDsp(),frm->GetWin()->XWin(),
											dsp->XDefaultRootWindow(),0,0,&xp,&yp,&dmy);
		frm->GetV(GT_Direction,&dir,GT_XSize,&xs,GT_YSize,&ys,GT_DONE);
		switch(dir)
		{
			case GTD_DOWN:
				this->MoveResize(xp,yp+ys,GTMAX(xsize,xs),ysize);
				GtPopup::Popup(xp,yp+ys,TRUE,tm);
				break;
			case GTD_UP:
				this->MoveResize(xp,yp-ysize,GTMAX(xsize,xs),ysize);
				GtPopup::Popup(xp,yp-ysize,TRUE,tm);
				break;
			case GTD_RIGHT:
				GtPopup::Popup(xp+xs,yp,TRUE,tm);
				break;
			case GTD_LEFT:
				GtPopup::Popup(xp-xsize,yp,TRUE,tm);
				break;
		}
	}
	else
	{
		XQueryPointer(dsp->XDsp(),dsp->XDefaultRootWindow(),&dmy,&dmy2,
									&xp,&yp,&xs,&ys,&dmsk);
		GtPopup::Popup(xp,yp,TRUE,tm);
	}
}

/************************************************************************\
*
* Popup menu without origin object.
*
\************************************************************************/
void GtMenu::Popup(ulong button,Time tm)
{
	this->Popup(NULL,button,tm);
}

/************************************************************************\
*
* Popdown menu and all sub menus.
*
\************************************************************************/
void GtMenu::Popdown(void)
{
	if(sub)
		sub->Popdown();
	GtPopup::Popdown();
	if(frm)
		frm->NotifyA(GT_Popdown,this);
}

/************************************************************************\
*
* Popup sub menu.
*
\************************************************************************/
void GtMenu::PopupSub(GtWidget *from,GtMenu *menu,Time tm)
{
	long xp,yp,xs,ys;
	
	sub=menu;
	from->GetV(GT_XPos,&xp,GT_YPos,&yp,GT_XSize,&xs,GT_YSize,&ys,GT_DONE);
	sxp=xp;
	syp=yp;
	sxs=xs;
	sys=ys;
	menu->Popup(from,but,tm);
}

/************************************************************************\
*
* Popdown sub menu.
*
\************************************************************************/
void GtMenu::PopdownSub(Time tm)
{
	if(sub)
	{
		gtime=tm;
		this->Grab();
		sub->Popdown();
	}
}

/************************************************************************\
*
* Handle mouse events.
*
\************************************************************************/
sbool GtMenu::Mouse(XEvent *xe)
{
	sbool ret;
	
	ret=GtPopup::Mouse(xe);
	if(xe)
	{
		switch(xe->type)
		{
			case ButtonRelease:
				if(xe->xbutton.button==but)
				{
					ret=TRUE;
					this->Popdown();
				}
				break;
		}
	}
	return(TRUE);
}
